@@ -4,15 +4,17 @@ use 5.006;
 use strict;
 use warnings;
 
-use Carp;
-use Carp::Assert;
-
-# List::SkipList::Node version is 0.04 - we cannot define it here in
-# one line because MakeMaker gets confused and pulls the first version
-# definition string it sees. (rt.cpan.org issues #4504).
+use Carp qw(carp croak);
+no Carp::Assert;
 
 our $VERSION
- = '0.04';
+ = '0.06';
+
+use constant NODEARGLIST => {
+  key    => 1,
+  value  => 1,
+  header => 1,
+};
 
 sub new {
   my $class = shift;
@@ -24,14 +26,24 @@ sub new {
   bless $self, $class;
 
   {
-    my %ARGLIST = ( map { $_ => 1 } qw( key value header ) );
     my %args = @_;
-    foreach my $arg_name (CORE::keys %args) {
-      if ($ARGLIST{$arg_name}) {
-	$self->$arg_name( $args{ $arg_name } );
-      } else {
-	croak "Invalid parameter name: ``$arg_name\'\'";
+
+    if (DEBUG) {
+      foreach my $arg_name (CORE::keys %args) {
+	if (NODEARGLIST->{$arg_name}) {
+	  $self->$arg_name( $args{ $arg_name } );
+	} else {
+	  croak "Invalid parameter name: ``$arg_name\'\'";
+	}
       }
+    } else {
+
+      # It's better to hard-code these in here and ignore invalid
+      # arguments for the speed improvement!
+
+      $self->key(    $args{key}    );
+      $self->value(  $args{value}  );
+      $self->header( $args{header} );
     }
   }
 
@@ -45,10 +57,9 @@ sub header {
   if (@_) {
     if (ref($_[0]) eq "ARRAY") {
       $self->{HEADER} = shift;
-      carp "Extra arguments ignored", if (@_);
+#      carp "Extra arguments ignored", if (@_);
     } else {
-      my @new_hdr = @_;
-      $self->{HEADER} = \@new_hdr;
+      $self->{HEADER} = \ @_;
     }
   } else {
     return wantarray ? @{$self->{HEADER}} : $self->{HEADER};
@@ -58,7 +69,7 @@ sub header {
 sub level {
   my $self = shift;
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-  return scalar @{$self->header()};
+  return scalar @{$self->{HEADER}};
 }
 
 sub forward
@@ -69,25 +80,23 @@ sub forward
     my $level = shift;
     assert (($level >= 0) ), if DEBUG;
 
-    my $hdr = $self->header;
-
     if (@_) {
       my $next = shift;
-      assert( !defined($next) ||
-	      UNIVERSAL::isa($next, __PACKAGE__) ), if DEBUG;
 
-      $hdr->[$level] = $next;
-      $self->header( $hdr );
+      assert( (!defined $next) || UNIVERSAL::isa($next, __PACKAGE__) ),
+	if DEBUG;
+
+      $self->{HEADER}->[$level] = $next;
 
     } else {
-      return $hdr->[$level];
+      return $self->{HEADER}->[$level];
     }
 
   }
 
 sub validate_key {
-#    my $self = shift;
-#    assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+  my $self = shift;
+  assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
   return 1;
 }
 
@@ -99,7 +108,7 @@ sub key {
     my $key = shift;
     assert( $self->validate_key( $key ) ), if DEBUG;
     $self->{KEY} = $key;
-    carp "Extra arguments ignored", if (@_);
+#    carp "Extra arguments ignored", if (@_);
   } else {
     return $self->{KEY};
   }
@@ -113,14 +122,14 @@ sub key_cmp {
   my $right = shift;
 
   assert( $self->validate_key( $right ) ), if DEBUG;
-  unless (defined $left) { return -1; }
+#  unless (defined $left) { return -1; } # we don't need this
 
   return ($left cmp $right);
 }
 
 sub validate_value {
-#    my $self = shift;
-#    assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+  my $self = shift;
+  assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
   return 1;
 }
 
@@ -132,7 +141,7 @@ sub value {
     my $value = shift;
     assert( $self->validate_value( $value ) ), if DEBUG;
     $self->{VALUE} = $value;
-    carp "Extra arguments ignored", if (@_);
+#    carp "Extra arguments ignored", if (@_);
   } else {
     return $self->{VALUE};
   }
@@ -144,38 +153,54 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = '0.33';
+our $VERSION = '0.40';
 
-use AutoLoader 'AUTOLOAD';
-use Carp;
-use Carp::Assert;
+use AutoLoader qw( AUTOLOAD );
+use Carp qw( carp croak );
+no Carp::Assert;
 
-use constant MAX_LEVEL => 32;
-use constant DEF_P     => 0.5;
+use constant MAX_LEVEL       => 32;
+use constant DEF_P           => 0.5;
+
+use constant BASE_NODE_CLASS => 'List::SkipList::Node';
+
+use constant ARGLIST => {
+  max_level  => 1,
+  p          => 1,
+  node_class => 1,
+};
+
+# Caching the "finger" for the last insert allows sequential (sorted)
+# inserts to be sped up.  It does not seem to affect the performance
+# of non-sequential inserts.
+
+use constant CACHE_INSERT_FINGERS => 1;
 
 sub new {
   no integer;
 
   my $class = shift;
 
-
   my $self = {
-    NODECLASS => 'List::SkipList::Node',
+    NODECLASS => BASE_NODE_CLASS,
     LIST      => undef,
     SIZE      => undef,
     MAXLEVEL  => MAX_LEVEL,
-    P         => DEF_P,
-    LASTNODE  => undef,                  # node with greatest key
-    LASTKEY   => undef,                  # last key used by next_key
+    P         => 0,
+    P_LEVELS  => [ ],
+    LASTNODE  => undef,                 # node with greatest key
+    LASTKEY   => undef,                 # last key used by next_key
+    LASTINSRT => undef,                 # used by CACHE_INSERTION_FINGERS
   };
 
   bless $self, $class;
 
+  $self->_set_p( DEF_P ); # initializes P_LEVELS
+
   {
-    my %ARGLIST = ( map { $_ => 1 } qw( max_level p node_class ) );
     my %args = @_;
     foreach my $arg_name (CORE::keys %args) {
-      if ($ARGLIST{$arg_name}) {
+      if (ARGLIST->{$arg_name}) {
 	my $method = "_set_" . $arg_name;
 	$self->$method( $args{ $arg_name } );
       } else {
@@ -218,7 +243,7 @@ sub clear {
   my $list = $self->_node_class->new ( header => [
      map { undef } (1..$self->_random_level)] );
 
-  assert( UNIVERSAL::isa($list, "List::SkipList::Node") ), if DEBUG;
+  assert( UNIVERSAL::isa($list, BASE_NODE_CLASS) ), if DEBUG;
 
   $self->{LIST}     = $list;
   $self->{SIZE}     = 0;
@@ -251,6 +276,24 @@ sub _set_p {
   assert( ($p>0) && ($p<1) ), if DEBUG;
 
   $self->{P} = $p;
+
+  # Because configuration is via hash, we may not set a new max_level
+  # before setting P, so we set to the standard MAX_LEVEL if it is
+  # greater.  Possible bug is if max_level is greater than MAX_LEVEL
+  # but is set after P.
+
+  my $n     = 1;
+  my $level = 0;
+  my $max   = (MAX_LEVEL > $self->max_level) ? MAX_LEVEL : $self->max_level;
+
+  $self->{P_LEVELS} = [ ]; 
+
+  while ($level <= $max) {
+    assert( $self->{P_LEVELS}->[$level-1] > $self->{P_LEVELS}->[$level] ),
+      if DEBUG;
+    $self->{P_LEVELS}->[$level++] = $n;
+    $n *= $p;
+  }
 }
 
 sub p {
@@ -279,63 +322,62 @@ sub level {
   return $self->list->level;
 }
 
+
 sub _random_level {
   no integer;
 
   my $self = shift;
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
+  my $n     = rand();
+  my $max   = $self->max_level;
   my $level = 1;
 
-  while ( (rand() < $self->p) && ($level < $self->max_level) ) {
-    $level ++; }
-
-  assert( ($level >= 1) && ($level <= $self->max_level) ), if DEBUG;
+  while ( ($level < $max) && ($n < $self->{P_LEVELS}->[$level]) ) {
+    $level++;
+  }
 
+  assert( ($level >= 1) && ($level <= $max) ), if DEBUG;
   return $level;
 }
 
 sub _search {
-  my $self = shift;
+  my $self   = shift;
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  my $list = $self->list;
+  my $list   = $self->list;
+  my $level  = $list->level-1;
 
-  my $key  = shift;
+  my $key    = shift;
 
-  my $x = $list;
+  my $finger = shift || [map { $list } (0..$level)];
+  assert( UNIVERSAL::isa($finger, "ARRAY") ), if DEBUG;
 
-  my @update;
-  my $finger = shift;
-  if (defined $finger) {
-    assert( UNIVERSAL::isa($finger, "ARRAY") ), if DEBUG;
-    @update = @{$finger};
-  } else {
-    @update = map { $list } (1..$list->level);
-  }
+  my $x      = $finger->[ $level ] || $list;
 
-  my $level     = $list->level-1;
+  # Iteresting Perl syntax quirk:
+  #   do { my $x = ... } while ($x)
+  # doesn't work because it considers $x out of scope.
+
+  my $cmp = 0;
 
   do {
-
-    $update[$level] = $x;
-
-    if ((!defined $x->forward($level)) ||
-	($x->forward($level)->key_cmp($key) > 0)) {
-
-      if ($level >= 0) { $level--; }
-
-    } else {
-      if ($x->forward($level)->key_cmp($key) <= 0) {
-	$x = $x->forward($level);	
+    my $fwd = $x->forward($level);
+       $cmp = (defined $fwd) ? $fwd->key_cmp($key) : 1; 
+    if ($cmp > 0) {
+      if ($level >= 0) {
+	$finger->[$level] = $x;
+	$level--;
       }
+    } else {
+      $finger->[$level] = $x;
+      $x = $fwd;
     }
+  } while (($level>=0) && ($cmp != 0));
 
-  } while (($level>=0) && ($x->key_cmp($key) != 0));
+  assert( UNIVERSAL::isa($x, BASE_NODE_CLASS) ), if DEBUG;
 
-  assert( UNIVERSAL::isa($x, "List::SkipList::Node") ), if DEBUG;
-
-  return ($x, \@update);
+  return ($x, $finger, $cmp);
 }
 
 sub insert {
@@ -344,55 +386,58 @@ sub insert {
 
   my $list   = $self->list;
 
-  my $key    = shift;
-  my $value  = shift;
+  my ($key, $value, $finger) = @_;
 
-  my $finger = shift;
+  # We save the node and finger of the last insertion. If the next key
+  # is larger, then we can use the "finger" to speed up insertions.
 
   if (defined $finger) {
     assert( UNIVERSAL::isa($finger, "ARRAY") ), if DEBUG;
+  } elsif (defined $self->{LASTINSRT}) {
+    if (CACHE_INSERT_FINGERS) {
+      my ($node, $ref) = @{ $self->{LASTINSRT} };
+    
+      if ($node->key_cmp($key) < 0) {
+	$finger = $ref;
+      }
+    }
   }
 
-  {
 
-    my ($x, $update_ref) = $self->_search($key, $finger);
+  my ($x, $update_ref, $cmp) = $self->_search($key, $finger);    
 
-    if ($x->key_cmp($key) == 0) {
-      $x->value($value);
-    } else {
+  if ($cmp == 0) {
+    $x->value($value);
+    $self->{LASTINSRT} = [$x, $update_ref],
+      if (CACHE_INSERT_FINGERS);
+  } else {
 
-      my $new_level = $self->_random_level;
+    my $new_level = $self->_random_level;
 
-      if ($new_level > $list->level) {
+    # We limit $new_level to no more than 1+$list-level. The speed
+    # improvement is quite noticeable.
 
-	for (my $i=$list->level; $i<$new_level; $i++) {
-	  $update_ref->[$i] = $list;
-	}
-      }
-
-      my $node = $self->_node_class->new( key => $key, value => $value );
-
-      for (my $i=0;$i<$new_level;$i++) {
-
-	if (defined $update_ref->[$i]->forward($i)) {
-	  $node->forward($i, $update_ref->[$i]->forward($i) );
-	} else {
-	  $node->forward($i, undef);
-	}
-	  
-	$update_ref->[$i]->forward($i,$node);
-      }
-
-      unless (defined $node->forward(0)) {
-	$self->{LASTNODE} = $node;
-      }
-
-      $self->{SIZE}++;
-      $self->reset;
+    if ($new_level > $list->level) {
+      $new_level  = 1 + $list->level;
+      $update_ref->[ $list->level ] = $list;
     }
 
-    return $update_ref;
+    my $node = $self->_node_class->new( key => $key, value => $value );
+
+    for (my $i=0;$i<$new_level;$i++) {
+      $node->forward($i, $update_ref->[$i]->forward($i) );	  
+      $update_ref->[$i]->forward($i,$node);
+    }
+
+    $self->{LASTNODE} = $node,
+      unless (defined $node->forward(0));
+
+    $self->{LASTINSRT} = [$node, $update_ref],
+      if (CACHE_INSERT_FINGERS);
+
+    $self->{SIZE}++;
   }
+  return $update_ref;
 }
 
 sub delete {
@@ -409,9 +454,9 @@ sub delete {
     assert( UNIVERSAL::isa($finger, "ARRAY") ), if DEBUG;
   }
 
-  my ($x, $update_ref) = $self->_search($key, $finger);
+  my ($x, $update_ref, $cmp) = $self->_search($key, $finger);
 
-  if ($x->key_cmp($key) == 0) {
+  if ($cmp == 0) {
     my $value = $x->value;
 
     my $level = $x->level; 
@@ -420,14 +465,19 @@ sub delete {
     for (my $i=0; $i<$level; $i++) {
 
       my $y = $update_ref->[$i];
-      while ($y->forward($i) != $x) {
-	$y = $y->forward($i);
-	assert( UNIVERSAL::isa($y, "List::SkipList::Node") ), if DEBUG;
+      while ((my $fwd = $y->forward($i)) != $x) {
+	$y = $fwd; # $y->forward($i);
+	assert( UNIVERSAL::isa($y, BASE_NODE_CLASS) ), if DEBUG;
       }
-      $y->forward($i, $x->forward($i));
- 
+      $y->forward($i, $x->forward($i)); 
     }
 
+    # There's probably a smarter way to handle this, but this is the
+    # safest way.\
+
+    $self->{LASTINSRT} = undef,
+      if (CACHE_INSERT_FINGERS);
+
     $self->{SIZE} --;
 
     undef $x;
@@ -435,8 +485,6 @@ sub delete {
     # it doesn't seem to be a wise idea to return a search finger for
     # deletions without further analysis
 
-    $self->reset;
-
     return $value;
 
   } else {
@@ -456,13 +504,11 @@ sub exists {
     assert( UNIVERSAL::isa($finger, "ARRAY") ), if DEBUG;
   }
 
-  my ($x, $update_ref) = $self->_search($key, $finger);
-
-  return ($x->key_cmp($key) == 0);
+  my ($x, $new_finger, $cmp) =  $self->_search($key, $finger);
+  return ($cmp == 0);
 }
 
 sub find {
-
   my $self = shift;
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
@@ -473,9 +519,9 @@ sub find {
     assert( UNIVERSAL::isa($finger, "ARRAY") ), if DEBUG;
   }
 
-  my ($x, $update_ref) = $self->_search($key, $finger);
+  my ($x, $update_ref, $cmp) = $self->_search($key, $finger);
 
-  if ($x->key_cmp($key) == 0) {
+  if ($cmp == 0) {
     return (wantarray)? ($x->value, $update_ref) : $x->value;
   } else {
     return;
@@ -499,9 +545,11 @@ sub first_key {
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
   my $list = $self->list;
-  if (defined $list->forward(0)) {
-    $self->{LASTKEY} = [$list->forward(0)->key, scalar $list->header];
-    $self->last_key;
+  my $fwd  = $list->forward(0);
+  if (defined $fwd) {
+    $self->{LASTKEY} =
+      [$fwd->key, [map { $list } (1..$list->level)] ];
+    return $self->last_key;
   } else {
     return;
   }
@@ -512,7 +560,7 @@ sub next_key {
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
   my $last_key = shift;
-  my $finger = shift;
+  my $finger   = shift;
 
   unless (defined $last_key) {
     ($last_key, $finger) = @{$self->{LASTKEY}};
@@ -523,10 +571,13 @@ sub next_key {
   }
 
   if (defined $last_key) {
-    my ($list, $update_ref) = $self->_search($last_key, $finger);
-    if ($list->key_cmp($last_key) == 0) {
-      if (defined $list->forward(0)) {
-	$self->{LASTKEY} = [$list->forward(0)->key, scalar $update_ref];
+    my ($list, $update_ref, $cmp) = $self->_search($last_key, $finger);
+
+    my $fwd  = $list->forward(0);
+ 
+    if ($cmp == 0) {
+      if (defined $fwd) {
+	$self->{LASTKEY} = [$fwd->key, $update_ref];
 	$self->last_key;
       } else {
 	return;
@@ -566,8 +617,9 @@ sub _first_node { # actually this is the second node
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
   my $list = $self->list;
-  if (defined $list->forward(0)) {
-    return ($list->forward(0), scalar $list->header);
+  my $fwd  = $list->forward(0);
+  if (defined $fwd) {
+    return ($fwd, scalar $list->header);
   } else {
     return;
   }
@@ -598,6 +650,12 @@ sub greatest {
   }
 }
 
+sub next {
+  my $self = shift;
+  assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+  die "unimplemented method";
+}
+
 sub keys {
   my $self = shift;
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
@@ -614,7 +672,6 @@ sub keys {
   return @keys;
 }
 
-
 sub values {
   my $self = shift;
   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
@@ -641,12 +698,15 @@ sub copy {
     node_class => $self->_node_class,
   );
 
-  my ($key, $finger) = $self->first_key;
+  my ($key, $finger_or) = $self->first_key();
+  my $value;
+  my $finger_cp = undef;
 
-  while (defined $key) {
-    $list->insert( $key, $self->find($key, $finger) );
-    ($key, $finger) = $self->next_key($key, $finger);
-  }
+  do {
+    ($value, $finger_or) = $self->find($key, $finger_or);
+    my $finger_cp        = $list->insert($key, $value, $finger_cp);
+    ($key, $finger_or)   = $self->next_key($key, $finger_or);
+  } while (defined $key);
 
   return $list;
 }
@@ -672,18 +732,18 @@ sub merge {
     
     if ($cmp < 0) {                     # key1 < key2
       if (defined $node1) {
-	$finger1 = $list1->insert( $node1->key, $node1->value, $finger1 );
+	$finger1 = $list1->insert( $node1->key, $node1->value, );
 	$node1 = $node1->forward(0);
       } else {
-	$finger1 = $list1->insert( $node2->key, $node2->value, $finger1 );
+	$finger1 = $list1->insert( $node2->key, $node2->value, );
 	$node2 = $node2->forward(0);
       }
     } elsif ($cmp > 0) {                # key1 > key2
       if (defined $node2) {
-	$finger1 = $list1->insert( $node2->key, $node2->value, $finger1 );
+	$finger1 = $list1->insert( $node2->key, $node2->value, );
 	$node2 = $node2->forward(0);
       } else {
-	$finger1 = $list1->insert( $node1->key, $node1->value, $finger1 );
+	$finger1 = $list1->insert( $node1->key, $node1->value, );
 	$node1 = $node1->forward(0);
       }
     } else {                            # key1 = key2
@@ -712,15 +772,19 @@ sub append {
     if ($list1->level > $list2->level) {
 
       if ($list1->level < $list1->max_level) {
-	$list1->list->forward($list1->level, $next);
+
+	my $i = $list1->level;
+	while (!defined $list1->list->forward($i)) { $i--; }
+	$list1->list->forward($i+1, $next);
       } else {
-	my $i = $list1->level -1;
+	my $i = $list1->level-1;
 	my $x = $list1->list->forward($i);
 	while (defined $x->forward($i)) {
 	  $x = $x->forward($i);
 	}
 	$x->forward($i, $next);
       }
+      $node->forward(0, $next);
 
     } else {
       for (my $i=0; $i<$node->level; $i++) {
@@ -750,12 +814,13 @@ sub _debug {
   my $list   = $self->list;
 
   while (defined $list) {
-    print $list->key||'undef', "=", $list->value||'undef'," ", $list,"\n";
+    print STDERR
+      $list->key||'undef', "=", $list->value||'undef'," ", $list,"\n";
 
     for(my $i=0; $i<$list->level; $i++) {
-      print " ", $i," ", $list->forward($i) || 'undef', "\n";
+      print STDERR " ", $i," ", $list->forward($i) || 'undef', "\n";
     }
-    print "\n";
+    print STDERR "\n";
 
     $list = $list->forward(0);
   }
@@ -804,12 +869,14 @@ A search would start at the top level: if the link to the right
 exceeds the target key, then it descends a level.
 
 Skip lists generally perform as well as balanced trees for searching
-but do not have the overhead with respect to inserting new items.
+but do not have the overhead with respect to inserting new items.  See
+the included file C<Benchmark> for a comparison of performance with
+other Perl modules.
 
 For more information on skip lists, see the L</"SEE ALSO"> section below.
 
-Only alphanumeric keys are supported.  To use numeric or other types
-of keys, see L</"Customizing the Node Class"> below.
+Only alphanumeric keys are supported "out of the box".  To use numeric
+or other types of keys, see L</"Customizing the Node Class"> below.
 
 =head2 Methods
 
@@ -1065,15 +1132,21 @@ developer use only.  These may change in future versions.
 
 =item _search
 
-  ($node, $header_ref) = $list->_search( $key );
+  ($node, $finger, $cmp) = $list->_search( $key );
 
 Searches for the node with a key.  If the key is found, that node is
 returned along with a L</"header">.  If the key is not found, the previous
 node from where the node would be if it existed is returned.
 
+Note that the value of C<$cmp>
+
+  $cmp = $node->key_cmp( $key )
+
+is returned because it is already determined by C<_search>.
+
 Search fingers may also be specified:
 
-  ($node, $header_ref) = $list->_search( $key, $finger );
+  ($node, $finger, $cmp) = $list->_search( $key, $finger );
 
 Note that the L</"header"> is actually a
 L<search finger|/"About Search Fingers">.
@@ -1337,6 +1410,12 @@ however, it may fail:
 
 Therefore, use search fingers with caution.
 
+Search fingers are specific to particular instances of a skip list.
+The following should not work:
+
+  ($value1, $finger) = $list1->find('bar');
+  $value2            = $list2->find('foo', $finger);
+
 One useful feature of fingers is with enumerating all keys using the
 C<first_key> and C<next_key> methods:
 
